# Copyright:
# Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

package MapWOCAnalyser::MapWOCAnalyser;

use strict;
use warnings;

use Exporter;
our @ISA = qw( Exporter );
our @EXPORT_OK = qw( @filename );

use Net::Pcap;
use NetPacket::Ethernet;
use NetPacket::IP qw( IP_PROTO_TCP );
use NetPacket::TCP;

use Digest::MD5 qw( md5 md5_hex );
use IO::Uncompress::Gunzip qw( gunzip $GunzipError );
use Cwd;
use XML::Simple;
use File::Basename;
use Digest::SHA1  qw(sha1_hex);
use Date::Calc qw(Normalize_DHMS Delta_DHMS);

#use Data::Dumper;

my $FALSE = 0;
my $TRUE = 1;

$/ = undef;
$\ = "";
$| = 0;

our @filename;

sub _new {

    my ( $this, %args ) = @_;
    my $self = {};

    my $err;

    bless $self, $this;

    $self->{ 'work_dir'    }  = $args{ 'work_dir'   };
    $self->{ 'BINARY_DIR'  }  = $args{ 'BINARY_DIR' };
    $self->{ 'LOG_DIR'     }  = $args{ 'LOG_DIR'    };
    $self->{ 'logfile'     }  = $args{ 'logfile'    };
    $self->{ 'logfileh'    }  = $args{ 'logfileh'   };
    $self->{ 'pcapfile'    }  = $args{ 'pcapfile'   };
    $self->{ 'xmlfile'     }  = $args{ 'xmlfile'    };

    # Read PCAP-file.
    my $pcap = Net::Pcap::open_offline(  $self->{ 'pcapfile' }, \$err );

    unless ( defined $pcap ) {
        print 'ERROR: Unable to create packet capture on file ',  $self->{ 'pcapfile' }, ' - ', $err . "\n";
        exit 255;
    }

    # Call syn_packets on all packets.
    Net::Pcap::loop( $pcap, -1, \&_syn_packets, $self ) ;
    Net::Pcap::close( $pcap );

    return $self;

}

sub _read_traffic {

    my $self = shift;
    my ( $infile ) = @_;

    my $xml;

    if( ! open( OFH, "<" . $infile ) ) {

        print  "ERROR: Could not open $infile for data evaluation.\n";
        return $FALSE;

    }

    binmode( OFH );
    seek   ( OFH, 0, 0 );
    my $data = <OFH>;
    close  ( OFH );

    my @data_split = split( /SPLITMWOC/, $data );

    foreach my $r ( @data_split ) {

        if( length( $r ) > 0 ) {

            my ( $request_header, $response_header, $response_body ) = split( /\n\n|\r\n\r\n/, $r ,3);

            if ( exists ( $self->{ basename ( $infile ) } ) ) {

                if( ! defined ( $self->{ 'pcap_start' } ) ) {

                    $self->{ 'pcap_start' } = $self->{ basename ( $infile ) }{ 'real_time' };

                    my $xml = '<time>' . $self->{ 'pcap_start' } . '</time>';

                    $self->_add_xml( XMLout ( $xml,
                            RootName      => undef,
                            NoAttr        => 0,
                            SuppressEmpty => 0,
                            NoSort        => 1,
                            NoEscape      => 0
                          )
                      );

                } else {

                    $self->{ 'pcap_stop' } = $self->{ basename ( $infile ) }{ 'real_time' };

                }

                if( defined ( $self->{ basename ( $infile ) }{ 'host' } && defined ( $self->{ basename ( $infile ) }{ 'fname' } ) ) ) {

                    $xml = {
                            url         => [ 'http://' . $self->{ basename ( $infile ) }{ 'host' } . '/' .  $self->{ basename ( $infile ) }{ 'fname' } ],
                            ip          => [ $self->{ basename ( $infile ) }{ 'ip' } ],
                            browser     => [ $self->{ basename ( $infile ) }{ 'browser' } ]
                      };

                    $self->_add_xml( XMLout ( $xml,
                            NoAttr        => 0,
                            RootName      => 'target',
                            SuppressEmpty => 1,
                            NoSort        => 1,
                            NoEscape      => 0
                          )
                      );

                }
            }

            if( $request_header =~ /HTTP\/1\.(1|0)/ && $response_header =~ /HTTP.1\.(1|0)/ ) {

                # HTTP Traffic.

                $self->_read_http( $request_header, $response_header, $response_body, $infile );
            }

            else {

                # Non HTTP Traffic.

                $self->_read_non_http( $r, $infile );
            }

        }
    }

    return $TRUE;

}

=item syn_packets()

# callback funktion for Net::Pcap::loop()

=cut

sub _syn_packets {

    my ($self, $header, $packet ) = @_;

    my ( $fname, $browser, $host, $referer );

    our @filename;

    # Strip ethernet encapsulation of captured packet.

    my $ether_dec  = NetPacket::Ethernet->decode( $packet );
    my $ether_data = NetPacket::Ethernet::strip( $packet );

   # Decode contents of TCP/IP packet contained within captured ethernet packet.

    my $ip  = NetPacket::IP->decode( $ether_data );

    return if( $ip->{ 'proto' } != IP_PROTO_TCP ); # ignore other than TCP (e.g. UDP)

    my $tcp = NetPacket::TCP->decode( $ip->{ 'data' } );

    my $peer_string =  $self->{ 'work_dir' } . "/" .$ip->{ 'src_ip' } . "_" . $tcp->{ 'src_port' } . "_" . $tcp->{ 'dest_port' };

    my $tcpdata = $tcp->{ 'data' };

    # Open File according to string that should represent this communication (it doesnt, as dst-port is not unique).
    if ( ( -e $peer_string ) && ( length ( $tcp->{ 'data' }) > 0 ) ) {

        open (AUS ,">>" . $peer_string);

    } else {

        $peer_string =  $self->{ 'work_dir' } . "/" .$ip->{ 'dest_ip' } . "_" . $tcp->{ 'dest_port' } . "_" . $tcp->{ 'src_port' };
        if ( ( -e $peer_string )  && ( length ( $tcp->{ 'data' }) > 0 ) ) {

            # What does this mean? It is the second packet of this peer combo?

            open ( AUS , ">>" . $peer_string );

        } else {

            if ( length ( $tcp->{'data'} ) >1 ) {

                open ( AUS , ">>" . $peer_string );
                push ( @filename, $peer_string );

            }
        }
    }

    if (length ( $tcpdata ) > 1) {

        my ( $sec, $min, $hour, $mday, $mon, $year, undef, undef, undef) = gmtime( $header->{ 'tv_sec' } );

        $mon++;
        $sec  = "0" . $sec  if ( $sec  < 10 );
        $min  = "0" . $min  if ( $min  < 10 );
        $hour = "0" . $hour if ( $hour < 10 );
        $mday = "0" . $mday if ( $mday < 10 );
        $mon  = "0" . $mon  if ( $mon  < 10 );

        my $real_time = $year + 1900 . "-" . $mon . "-" . ${mday} ."-" . $hour .":" . $min . ":" . $sec;

        $self->{ basename ( $peer_string ) }{ 'time'      } = $header->{ 'tv_sec' };
        $self->{ basename ( $peer_string ) }{ 'real_time' } = $real_time;

        if (  $tcpdata =~ /^(GET|POST)/ ) {

            print AUS "SPLITMWOC";
            print AUS "TIME:" . $header->{ 'tv_sec' } . "\n";

            my $h1 = substr( $tcpdata, 0, 500 );

            if( defined $h1 ){

                $h1 =~ s/(GET|POST) \/(.*) HTTP\/1\../$fname = $2;/egi;
                $h1 =~ s/Referer: (http[s]*:\/\/[^\n]+)/$referer = $1;/eg;
                $h1 =~ s/Host: ([^\n]+)./$host = $1;/eg;
                $h1 =~ s/User-Agent: ([^\n]+)/$browser = substr( $1, 0, 12 );/eg;

            }

            if( ! ( defined $browser ) ) { $browser = "NODEF"; }
            if( ! ( defined $fname   ) ) { $fname   = "";}

            $self->{ basename ( $peer_string ) }{ 'ip'        } = $ip->{ 'dest_ip' };
            $self->{ basename ( $peer_string ) }{ 'host'      } = $host    ? $host    : "NO_HOST";
            $self->{ basename ( $peer_string ) }{ 'fname'     } = $fname   ? $fname   : "NO_FNAME";
            $self->{ basename ( $peer_string ) }{ 'referer'   } = $referer ? $referer : "NO_REFERER";
            $self->{ basename ( $peer_string ) }{ 'browser'   } = $browser;

            if( ! ( defined $referer ) && ( defined $host ) ) {

                open(NFH, ">>".$self->{ 'logfileh' } ) or die " Could not open $self->{ 'logfileh' }\n";
                print NFH $ip->{'src_ip'} . "\t" . $header->{ 'tv_sec' }. "\t$host/$fname\t$browser\n";
                close(NFH);

            }

        }

        print AUS  "$tcpdata";
        close(AUS);

    }
}

sub _read_http {

    my $self = shift;

    my ($request_header, $response_header, $response_body, $infile) = @_;
    my ($request, $requestfile);

    if ( ( defined $response_body ) && ( $response_body =~ /^\x1f\x8b\x08/ ) ) {
        my $uncompress;
        gunzip \$response_body => \$uncompress;
        $response_body = $uncompress;
    }

    if( defined $response_body ) {

        my $f_type = $self->_file_type( substr( $response_body, 0, 300 ) );

        if( defined ( $f_type ) && $self->_is_interesting_filetype( $f_type ) ) {

            my ( $outfile1, $logdata, $filename ) = $self->_make_outfile1( $response_header, $request_header, $infile );

            if( ! defined ( $outfile1 ) || ! defined ( $logdata ) ) {
                return $FALSE;
            }

            my $md   = md5_hex(  $response_body );
            my $sha1 = sha1_hex( $response_body );

            if ( open (DA, ">> $self->{ 'BINARY_DIR' }/" .  $outfile1 . "-" . $md ) ) {

                binmode(DA);
                print DA "$response_body";
                close(DA);

                my $data2 = `file -b $self->{ 'BINARY_DIR' }/$outfile1-$md`;
                $data2 =~ s/[\r\n]+//g;

                my $mime_typ = `file -bi $self->{ 'BINARY_DIR' }/$outfile1-$md`;
                $mime_typ =~ s/[\r\n]+//g;

                my $ssdeep = $self->_ssdeep ( $self->{ 'BINARY_DIR' } . "/" . $outfile1 . "-" . $md );

                my $erg = $self->_get_first_html( $logdata );

                open(NFH, ">>".$self->{ 'logfile' }) or die "could not open $self->{ 'logfile' }\n";
                binmode(NFH);
                if( ! defined $erg ) { $erg = "NO_ERG"; }

                print NFH "[" . uc( $f_type )  . "]\t$logdata\t==>$data2<==\t$md\t$self->{ 'pcapfile' }\t$erg\n";
                close(NFH);

                ( $request, $requestfile ) = $self->_parse_request_header( $request_header );
                my $file_type = $self->_get_filetypestring( $f_type );

                my $xml = {

                    file => {

                        source      => [ $request ],
                        filename    => [ $requestfile ],
                        extractname => [ $self->{ 'BINARY_DIR' } . "/" .  $outfile1 . "-" . $md ],
                        analysis    => {

                            method => [

                                {
                                    name   =>  "mapwocanalyse.pl/type_classifier",
                                    output => [ $file_type ]
                                },

                                {
                                    name   =>  "tool:file -b",
                                    output => [ $data2 ]
                                },

                                {
                                    name   =>  "tool:file -bi",
                                    output => [ $mime_typ ]
                                },

                                {
                                    name   =>  "tool:sha1sum",
                                    output => [ $sha1 ]
                                },

                                {
                                    name   =>  "tool:ssdeep",
                                    output => [ $ssdeep ]
                                },

                                {
                                    name   =>  "tool:clamav",
                                    output => [ "###" . $outfile1 . "-" . $md . "###" ]
                                }

                              ]

                          }
                      },

                    dumpfile => {

                        file     =>  [ $infile ],
                        analysis => {

                            method => [

                                {
                                    name   =>  "tool:clamav",
                                    output => [ "###" . basename ( $infile ) . "###" ]
                                }

                              ]
                          }

                      }

                  };

                $self->_add_xml( XMLout ( $xml,
                        RootName      => "files",
                        NoAttr        => 0,
                        SuppressEmpty => 0,
                        NoSort        => 1,
                        NoEscape      => 0
                      )
                  );

            }

        }
    }

    if( defined ( $response_header ) && defined ( $request_header ) ) {

        $self->_log_http_code( $response_header, $request_header, $infile, 301 );
        $self->_log_http_code( $response_header, $request_header, $infile, 302 );
        $self->_log_http_code( $response_header, $request_header, $infile, 303 );
        $self->_log_http_code( $response_header, $request_header, $infile, 307 );
        $self->_log_http_code( $response_header, $request_header, $infile, 404 );

        my ( $return_code, $loc ) = $self->_get_http_return_code( $response_header );

        my $xml = {

            connection => {

                sequence  => $self->{ basename ( $infile ) }{ 'time' },
                time      => $self->{ basename ( $infile ) }{ 'real_time' },
                protocol  => [ "HTTP" ],
                ressource => [ $request ],
                code      => [ $return_code ],
                ip        => [ $self->{ basename ( $infile ) }{ 'ip' } ],
                location  => [ $loc ],
                findings  => {

                  }
              }
          };

        if( defined ( $response_body ) ) {

            if ( my $google_referer = $self->_strip_google( $response_header, $request_header, $response_body, $infile ) ) {

                $xml->{ 'connection'  }{ 'findings' } = { method => [
                        {
                            name    => "GoogleAnalyticsParser",
                            output  => [ "Referrer: " . $google_referer ]
                        }
                      ]
                  };
            }

            if ( $self->_strip_javascript( $response_header, $request_header, $response_body, $infile ) ) {

                $xml->{ 'connection'  }{ 'findings' } = { method => [
                        {
                            name    => "JSAnalyser",
                            output  => [ "JavaScript found" ]
                        }
                      ]
                  };

            }

            if ( $self->_strip_google_safebrowsing( $response_header, $request_header, $response_body, $infile ) ) {

                $xml->{ 'connection'  }{ 'findings' } = { method => [
                        {
                            name    => "GoogleSafebrowsingParser",
                            output  => [ "Google Safebrowsing found" ]
                        }
                      ]
                  };

            }

        }

        $self->_add_xml( XMLout ( $xml,
                RootName      => "connections",
                NoAttr        => 0,
                SuppressEmpty => 1,
                NoSort        => 1,
                NoEscape      => 0
              )
          );

    }

}

sub _read_non_http {

    my $self = shift;
    my ($r, $infile) = @_;

    my $f_type = $self->_file_type( substr( $r, 0, 300 ) );

    if( defined ( $f_type ) &&  $self->_is_interesting_filetype( $f_type ) ) {

        if ($r =~ /^\x1f\x8b\x08/ )  {
            my $uncompress;
            gunzip \$r => \$uncompress;
            $r = $uncompress;
        }

        my $md   = md5_hex(  $r );
        my $sha1 = sha1_hex( $r );

        if ( open (DA, ">>$self->{ 'BINARY_DIR' }/NON_HTTP_dump-" . $md ) ) {
            binmode(DA);
            print DA "$r";
            close(DA);

        }

        my $data2 = `file -b $self->{ 'BINARY_DIR' }/NON_HTTP_dump-$md`;
        $data2 =~ s/[\r\n]+//g;

        my $mime_typ = `file -bi $self->{ 'BINARY_DIR' }/NON_HTTP_dump-$md`;
        $mime_typ =~ s/[\r\n]+//g;

        my $ssdeep = $self->_ssdeep (  $self->{ 'BINARY_DIR'} . '/NON_HTTP_dump-' . $md );

        my $logdata = $infile . "\t0\tNO_HOST\tNO_FNAME\tNO_REFERER\tNO_BROWSER\n";

        my $erg = $self->_get_first_html( $logdata );

        open(NFH, ">>".$self->{ 'logfile' }) or die "could not open $self->{ 'logfile' }\n";
        binmode(NFH);
        if( ! defined $erg ) { $erg = "NO_ERG"; }

        print NFH "[" . uc( $f_type )  . "] $logdata\t==>$data2<==\t$md\t$self->{ 'pcapfile' }\t$erg\n";
        close(NFH);

        my $file_type = $self->_get_filetypestring( $f_type );
        my ( $traffic_type, $traffic_file, $traffic_source ) = $self->_guess_traffic_type( substr( $r, 0, 800 ) );

        my $xml = {

            file => {

                source      => [ $traffic_source ],
                filename    => [ $traffic_file ],
                extractname => [ $self->{ 'BINARY_DIR' } . "/NON_HTTP_dump-" . $md ],
                analysis    => {

                    method => [

                        {
                            name   =>  "mapwocanalyse.pl/type_classifier",
                            output => [ $file_type ]
                        },

                        {
                            name   =>  "tool:file -b",
                            output => [ $data2 ]
                        },

                        {
                            name   =>  "tool:file -bi",
                            output => [ $mime_typ ]
                        },

                        {
                            name   =>  "tool:sha1sum",
                            output => [ $sha1 ]
                        },

                        {
                            name   =>  "tool:ssdeep",
                            output => [ $ssdeep ]
                        },

                        {
                            name   =>  "tool:clamav",
                            output => [ "###NON_HTTP_dump-" . $md . "###" ]
                        }

                      ]

                  }
              },

            dumpfile => {

                file     =>  [ $infile ],
                analysis => {

                    method => [

                        {
                            name   =>  "tool:clamav",
                            output => [ "###" . basename ( $infile  ) . "###" ]
                        }

                      ]
                  }

              }
          };

        $self->_add_xml( XMLout ( $xml,
                RootName      => "files",
                NoAttr        => 0,
                SuppressEmpty => 0,
                NoSort        => 1,
                NoEscape      => 0
              )
          );

        $xml = {

            connection => {

                sequence  => $self->{ basename ( $infile ) }{ 'time' },
                time      => $self->{ basename ( $infile ) }{ 'real_time' },
                protocol  => [ $traffic_type ],

                #                            ressource => [ $request ],
                code      => [ "none" ],
                ip        => [ $self->{ basename ( $infile ) }{ 'ip' } ],
                findings  => {

                  }
              }
          };

        $self->_add_xml( XMLout ( $xml,
                RootName      => "connections",
                NoAttr        => 0,
                SuppressEmpty => 1,
                NoSort        => 1,
                NoEscape      => 0
              )
          )

    }

}

=item file_type()

# File_Type - return file typ for given data, else just "data".

=cut

sub _file_type {

    my $self = shift;
    my $data = shift;

    if( length( $data ) < 1 ) {
        return "empty_body";
    }

    my $type = "";

    if ($data =~ /^GIF8[7-9]/)                                                            { $type = "gif";    }
    elsif ($data =~ /^\377.....(JPEG|JFIF)/)                                              { $type = "jpeg";   }
    elsif ($data =~ /^.PNG/)                                                              { $type = "png";    }
    elsif ($data =~ /^PK\003\004/)                                                        { $type = "zip";    }
    elsif ($data =~ /^\%PDF/)                                                             { $type = "pdf";    }
    elsif ($data =~ /^\037\213/)                                                          { $type = "gz";     }
    elsif ($data =~ /^BZh/)                                                               { $type = "bz2";    }
    elsif ($data =~ /^\177ELF/)                                                           { $type = "elf";    }
    elsif ($data =~ /^\%!/)                                                               { $type = "ps";     }
    elsif ($data =~ /ustar/)                                                              { $type = "tar";    }
    elsif ($data =~ /^\x4D\x5A.\x00.\x00\x00\x00.\x00\x00\x00\xff\xff\x00\x00/)           { $type = "exe";    }
    elsif ($data =~ /^\xca\xfe\xba\xbe\x00\x00\x00/)                                      { $type = "mach-o"; }
    elsif ($data =~ /<html>/i)                                                            { $type = "html";   }
    elsif ($data =~ /^\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1/)                                  { $type = "ms_doc"; }
    else { $type = "data"; }

    return $type;
}

=item is_interesting_filetype()

# is_interesting_filetype: check if file typ is interesting for us

=cut

sub _is_interesting_filetype {

    my $self = shift;
    my $file_type = shift;

    return $TRUE; # atm we are interested in all filetypes.

    if( $self->_is_image( $file_type ) ) { # If we are not interested in images.
        return $FALSE;
    }

    my %intersting;

    $intersting{ 'pdf' }     = 1;
    $intersting{ 'exe' }     = 1;
    $intersting{ 'tar' }     = 1;
    $intersting{ 'elf' }     = 1;
    $intersting{ 'zip' }     = 1;
    $intersting{ 'ms_doc'}   = 1;
    $intersting{ 'mach-o' }  = 1;
    $intersting{ 'gz' }      = 1;
    $intersting{ 'bz2' }     = 1;

    if( exists ( $intersting{ $file_type } ) ) {
        return $TRUE;
    }

    return $FALSE;

}

sub _get_filetypestring {

    my $self = shift;
    my $file_type = shift;

    my %type;

    $type{ 'pdf'    }  = "Portable Document Format";
    $type{ 'exe'    }  = "Windows Executable/PE";
    $type{ 'tar'    }  = "Tape Archive";
    $type{ 'elf'    }  = "Executable and Linkable Format";
    $type{ 'zip'    }  = "Compressed ZIP";
    $type{ 'ms_doc' }  = "MS Office Dokument";
    $type{ 'mach-o' }  = "Mach object file";
    $type{ 'gz'     }  = "GNU zip";
    $type{ 'bz2'    }  = "Compressed Bzip2";
    $type{ 'gif'    }  = "Graphics Interchange Format";
    $type{ 'png'    }  = "Portable Network Graphics";
    $type{ 'jpeg'   }  = "JPEG";
    $type{ 'html'   }  = "HTML";

    if ( exists ( $type{ $file_type } ) ) {

        return $type{ $file_type };

    } else {

        return "unknnown filetype";

    }

}

=item is_image()

=cut

sub _is_image {

    my $self = shift;
    my $file_type = shift;

    if( $file_type eq "gif" || $file_type eq "png" || $file_type eq "jpeg" ) {
        return $TRUE;
    }

    return $FALSE;

}

=item log_http_code()

=cut

sub _log_http_code {

    my $self = shift;
    my ($response_header, $request_header, $infile, $http_code) = @_;

    if ( ( $response_header =~ /HTTP\/1\.(0|1) $http_code/ ) ) {

        my ( $outfile1, $logdata ) = $self->_make_outfile1( $response_header, $request_header, $infile );

        if( ! defined ( $outfile1 ) || ! defined ( $logdata ) ) {
            return $FALSE;
        }

        my $erg = $self->_get_first_html( $logdata );
        if( ! defined ( $erg ) ) {
            $erg = "NO_ERG";
        }

        open(NFH, ">>" .  $self->{ 'LOG_DIR' } . "/" . $http_code . ".log") or die "could not open $self->{ 'LOG_DIR' } . "/" . $http_code\n" ;
        binmode(NFH);
        print NFH "[$http_code]\t$logdata\t$erg\n";
        close(NFH);

        return $TRUE;
    }

    return $FALSE;

}

sub _get_http_return_code {

    my $self = shift;
    my $response_header = shift;

    my ($http_code, $location);

    $response_header =~ s/HTTP\/1\.(0|1) ([0-9]+)/$http_code = $2;/eg;
    $response_header =~ s/Location: ([^\n]+)/$location = $1;/eg;

    if ( ! defined ( $location ) ) {
        $location = undef;
    } else {
        $location =~ s/\x0D//;
        $self->{ 'location' } = $location;
    }

    if ( ! defined ( $http_code ) ) {
        $http_code = "NO_HTTP_CODE";
    }

    return ($http_code, $location);

}

=item strip_javascript()

=cut

sub _strip_javascript {

    my $self = shift;
    my ( $response_header, $request_header, $response_body, $infile ) = @_;

    if ( $response_header =~ /[jJ]avascript/) {

        my ( $outfile1, $logdata ) = $self->_make_outfile1( $response_header, $request_header, $infile );

        if( ! defined ( $outfile1 ) || ! defined ( $logdata ) ) {
            return $FALSE;
        }

        if ( open(NFH, "> " . $self->{ 'BINARY_DIR' } . "/" . "$outfile1") ) {
            binmode(NFH);
            print NFH "$response_body";
            close(NFH);
        }

        my $erg = $self->_get_first_html( $logdata );

        if( ! defined ( $erg ) ) {
            $erg = "NO_ERG";
        }

        open(NFH, ">>".$self->{ 'logfile' }) or die "Could not open $self->{ 'logfile' }\n" ;
        binmode(NFH);
        print NFH "[Javascript]\t$logdata\t$erg\n";
        close(NFH);

        return $TRUE;
    }

    return $FALSE;
}

=item strip_google()

=cut

sub _strip_google {

    my $self = shift;
    my ( $response_header, $request_header, $response_body, $infile ) = @_;

    my $header1 = $request_header . $response_header;
    my ( $logheader, $ftime, $fname, $referer, $host, $browser, $location, $outfile1, $logdata );

    if( defined $header1 ) {

        $header1=~s/TIME:(.[a-zA-Z0-9 :]{1,})/$ftime=$1;/eg;

        if( defined $ftime ) {

            $header1=~s/Referer: http:\/\/(.[\x10-\xff]{1,})[\x0a\x0d]/$referer=$1;/eg;
            $header1=~s/Host: (.[\x10-\xff]{1,})[\x0a\x0d]/$host=$1;/eg;

            if( ! defined $referer ) { $referer = "NO_REFERER"; }
            if( ! defined $host )    { $host    = "NO_HOST"; }

            $logheader =$infile."\t" . $ftime . "\tHost:" . $host . "\tReferer:" . $referer . "\n";

        }

        if( ( defined $host ) && ( $host =~ /[Gg][Oo][Oo][Gg][Ll][Ee]-[Aa][Nn][Aa]/) ) {

            my ( $outfile1, $logdata ) = $self->_make_outfile1( $response_header, $request_header, $infile );

            if( ! defined ( $outfile1 ) || ! defined ( $logdata ) ) {
                return $FALSE;
            }

            my $md = md5_hex( $response_body );

#            open(NFH, ">Google-" . $outfile1 . $md )or die "Could not open Google-$outfile1$md\n";
#            binmode(NFH);
#            print NFH $response_body;
#            close(NFH);

            #my $erg = get_first_html($logdata);

            open(NFH, ">>".$self->{ 'logfile' }) or die "Could not open $self->{ 'logfile' }\n" ;
            binmode(NFH);
            print NFH "[GOOGLE]\t$logheader\n";
            close(NFH);

            open(NFH, "> $self->{ 'BINARY_DIR' }/GOOGLE-" . $outfile1 . "-" .$md ) or die "Could not open $outfile1";
            binmode(NFH);
            print NFH "$response_body";
            close(NFH);

            return $referer;
        }

        return $FALSE;
    }

    return $FALSE;

}

=item strip_google_safebrowsing()

=cut

sub _strip_google_safebrowsing {

    my $self = shift;
    my ( $response_header, $request_header, $response_body, $infile ) = @_;

    if ( ( $response_header =~ /safebrowsing/ ) && ( $response_header =~ /gethash/ ) ) {

        my ( $outfile1, $logdata ) = $self->_make_outfile1( $response_header, $request_header, $infile );

        if( ! defined ( $outfile1 ) || ! defined ( $logdata ) ) {
            return $FALSE;
        }

        if ( open(NFH, "> $self->{ 'BINARY_DIR' }/"."$outfile1") ) {
            binmode(NFH);
            print NFH "$response_header\n$response_body";
            close(NFH);
        }

        my $erg = $self->_get_first_html( $logdata );

        open(NFH, ">>".$self->{ 'logfile' }) or die "could not open $self->{ 'logfile' }\n" ;
        binmode(NFH);
        print NFH "[Google_Safebrowsing]\t$logdata\t$erg\n";
        close(NFH);

        return $TRUE;

    }

    return $FALSE;

}

=item make_outfile1()

=cut

sub _make_outfile1 {

    my $self = shift;
    my ($response_header, $request_header, $infile) = @_;

    my $header = $request_header . $response_header;

    my ($ftime, $fname, $referer, $host, $browser, $location, $outfile1, $logdata);

    if( defined $header ) {

        $header =~ s/TIME:(.[a-zA-Z0-9 :]{1,})/$ftime = $1;/eg;
        $header =~ s/(GET|POST)[\x00-\xff]{1,}\/(.[\x00-\xff]{1,}) HTTP\/1\..[\x0a\x0d]/$fname = $2;/eg;
        $header =~ s/Referer: http:\/\/(.[\x10-\xff]{1,})[\x0a\x0d]/$referer = $1;/eg;
        $header =~ s/Host: (.[\x10-\xff]{1,})[\x0a\x0d]/$host = $1;/eg;
        $header =~ s/User-Agent: (.[\x10-\xff]{1,})[\x0a\x0d]/$browser = $1;/eg;

        $header =~ s/Location: ([^\n]+)/$location = $1;/eg;

        if( ! defined $ftime)   { $ftime   = 0;            }
        if( ! defined $fname)   { $fname   = "NO_NAME";    }
        if( ! defined $referer) { $referer = "NO_REFERER"; }
        if( ! defined $host)    { $host    = "NO_HOST";    }
        if( ! defined $browser) { $browser = "NO_BROWSER"; }

        $logdata = $infile."	".$ftime."	".$host."	".$fname."	".$referer."	".$browser;

        if( defined ( $location ) ) {
            $location =~ s/[\r\n]+//g;
            $logdata .= "      " . $location;
        }

        if (length $host    > 15)  {   $host    = substr( $host, 0, 13 );    }
        if (length $referer > 15)  {   $referer = substr( $referer, 0, 13 ); }
        if (length $fname   > 15)  {   $fname   = substr( $fname, 0, 13 );   }

        $outfile1 = $host . "_" . $referer . "_" . $fname;
        $outfile1 =~ s/[\?&;=*"'\/!'`><|]/_/mg;

    }
    else {
        my $rand = int( rand( 1000000 ) );
        $outfile1="NO_HEADERINFO_" . $rand;
    }

    if( defined ( $outfile1 ) && defined ( $logdata ) ) {
        return ( $outfile1, $logdata, $fname );
    } else {
        return $FALSE;
    }

}

=item get_first_html()

=cut

sub _get_first_html {

    my $self = shift;
    my $logdata = shift;

    my $erg;
    if (defined $logdata) {

        # Split log data at tabs.
        my @ldata = split(/\x09/,$logdata);
        my $ip_adr;
        $ldata[0]=~ s/\/([0-9.]{1,})_/$ip_adr=$1;/eg;
        if( !defined $ldata[1]) { $ldata[1]=0;}
        my $ti=$ldata[1];
        my $brow=substr($ldata[5],0,12);

        # Set separator.
        $\="\n";
        open(INF, "<".$self->{ 'logfileh' }) or die " Could not open $self->{ 'logfileh' }\n";
        my $endflag=0;
        my $line;
        while ($endflag==0 && !eof(INF)){
            $line=readline (INF);
            my @gdata = split(/\n/,$line);
            my $glength=@gdata-1 ;
            do{
                my @hdata = split(/\x09/,$gdata[$glength]);
                if ( ($ip_adr eq $hdata[0]) && ($endflag != 1) ){
                    if ( (($ti-120) <= $hdata[1])  && ($ti >=$hdata[1]) ) {
                        if (( $brow eq $hdata[3]) &&( !($hdata[2] =~ /safebrowsing/))) {
                            $erg=$hdata[2];
                            $endflag=1;
                        }
                    }
                }
                $glength--;
              } while (($glength ne 0));
            close(INF);

            # Unset separator.
            $\="";
        }
        return $erg;
    }
}

sub _start_xml {

    my $self = shift;
    my $hcsid = shift;

    my $xml = '<?xml version="1.0"?>' . "\n"  . '<visit visitor="' . $hcsid  . '">' . "\n";
    $self->{'current_xml'} = $xml;

    return $TRUE;
}

sub _close_xml {

    my $self = shift;

    my $duration = "unknown";

    if( defined ( $self->{ 'pcap_start' } ) && defined ( $self->{ 'pcap_stop' } ) ) {

        my $start = $self->{ 'pcap_start' };
        my $stop  = $self->{ 'pcap_stop' };

        $start =~ s/:/-/g;
        $stop  =~ s/:/-/g;

        my @start = split( /-/, $start );
        my @stop  = split( /-/, $stop  );

        my ( undef, $Dh, $Dm, $Ds) = Delta_DHMS( $start[0], $start[1], $start[2], $start[3], $start[4], $start[5], $stop[0], $stop[1], $stop[2], $stop[3], $stop[4], $stop[5]);

        if( $Dh < 10 ) { $Dh = "0" . $Dh; }
        if( $Dm < 10 ) { $Dm = "0" . $Dm; }
        if( $Ds < 10 ) { $Ds = "0" . $Ds; }

        $duration = $Dh . ":" . $Dm . ":" . $Ds;

    }

    my $xml = "<duration>" . $duration . "</duration>\n";

    $xml .= "</visit>";
    $self->{'current_xml'} .= $xml;

    if( ! open( XML, ">> $self->{ 'xmlfile' }" ) ) {
        print "ERROR: Could not open XML-File $self->{ 'xmlfile' } for writing in Function close_xml(). Exit now!\n";
        exit 255;
    }

    print XML "$self->{'current_xml'}";
    close( XML );

    return $TRUE;

}

sub _add_xml {

    my $self = shift;
    my $xml = "@_";
    $self->{'current_xml'} .= $xml;

    return $TRUE;

}

sub _add_antivir_result_to_xml {

    my $self = shift;
    my @antivir_output = @_;

    my $antivir_file = basename( $antivir_output[0] );
    $antivir_file =~ s/:$//g;

    my $antivir_result = $antivir_output[1];
    $self->{'current_xml'} =~ s/###$antivir_file###/$antivir_result/g;

    return $TRUE;

}

sub _clean_xml {

    my $self = shift;

    $self->{'current_xml'} =~ s/###(.*)###/no virus/g;

    return $TRUE;
}

sub _parse_request_header {

    my $self   = shift;
    my $header = shift;

    my ( $host, $fname, $request );

    if( $header =~ m/(GET|POST) (.*) HTTP\/1\..[\x0a\x0d]/i ) {
        $request = $2;
    }

    $header =~ s/(GET|POST)[\x00-\xff]{1,}\/(.[\x00-\xff]{1,}) HTTP\/1\..[\x0a\x0d]/$fname = $2;/egi;
    $header =~ s/Host: (.[\x10-\xff]{1,})[\x0a\x0d]/$host = $1;/eg;

    if( ! defined ( $request ) ) { $request = "NO_REQUEST"; }
    if( ! defined ( $fname   ) ) { $fname   = "NO_FNAME"; }
    if( ! defined ( $host    ) ) { $host    = "NO_HOST"; }

    return ( "http://" . $host . $request, $fname )

}

sub _ssdeep {

    my $self = shift;
    my $file = shift;

    my $ssdeep = `ssdeep -sb $file 2>/dev/null`;
    my @sd = split(/\n/, $ssdeep);

    if ( defined ( $sd[1] ) ) {

        return $sd[1];

    }  else {

        return "NO SSDEEP";

    }

}

sub _guess_traffic_type { # At the moment only FTP

    my $self    = shift;
    my $traffic = shift;

    my $traffic_file   = "UNKNOWN";
    my $traffic_type   = "NON_HTTP/UNKNOWN";
    my $traffic_source = "UNKNOWN";

    if( defined ( $self->{ 'next_is_ftp_data' }  ) ) {

        $self->{ 'next_is_ftp_data' } = undef;

        $traffic_type = "FTP";

        if ( defined ( $self->{ 'location' } ) ) {

            if( $self->{ 'location' } =~ m/^ftp:\/\//i ) {
                $traffic_source = $self->{ 'location' };
            }

            $self->{ 'location' } = undef;
        }

        if( defined ( $self->{ 'next_ftp_data_trafficfile' } ) ) {

            $traffic_file = $self->{ 'next_ftp_data_trafficfile' };
            $self->{ 'next_ftp_data_trafficfile' } = undef;

        }
    }

    if ( $traffic =~ m/^220(.*)FTP/i ) {

        $traffic_type = "FTP";
        $self->{ 'next_is_ftp_data' } = 1;

        if ( $traffic =~ m/(get|recv|retr)\s([^\n]+)/i ) {

            $traffic_file = $2;
            $traffic_file =~ s/\x0D//;
            $self->{ 'next_ftp_data_trafficfile' } = $traffic_file;
            $traffic_file = "FTP-Header";

            if ( defined ( $self->{ 'location' } ) ) {

                if( $self->{ 'location' } =~ m/^ftp:\/\//i ) {

                    $traffic_source = $self->{ 'location' };

                }

            }

        }
    }

    return ( $traffic_type, $traffic_file, $traffic_source );

}

1;

